#include <queue>
#include <algorithm>
#include <iostream>
#include <string>

template <typename T, typename Op = std::less<T> >
class stable_priority_queue
{
	protected:
        template <typename ElemT, typename ElemOp>
		struct Element {
 			ElemT elem;
 			unsigned num;
 			const ElemOp *op;

 			Element(const ElemT& t, const ElemOp *_op, unsigned c)
	 			: elem(t), num(c), op(_op) {}

 			bool operator < (const Element<ElemT, ElemOp>& e1) const
 			{
	 			if (!((*op)(elem, e1.elem)) && !((*op)(e1.elem, elem))) {
		 			return num > e1.num;
	 			} else
		 			return ((*op)(elem, e1.elem));
 			}
		};

		typedef std::priority_queue<Element<T,Op> > PriorityQueue;

	protected:
	    PriorityQueue pqueue;
		Op op;
		unsigned counter;

	public:
		stable_priority_queue(Op _op = Op())
			: op(_op), counter(0)
		{
		}

		void push(const T& t)
		{
			pqueue.push(Element<T,Op>(t, &op, ++counter));
		}

		const T& top() const
		{
			return pqueue.top().elem;
		}

		void pop()
		{
			pqueue.pop();
		}

		bool empty() const
		{
			return pqueue.empty();
		}
};

struct X {

	int p;
	std::string s;

	bool operator < (const X& x) const {
		return p < x.p;
	}

	X(const char* _s, int _p)
		: p(_p), s(_s)
	{
	}
};

std::ostream& operator << (std::ostream& o, const X& x) {
	o << x.p << ": " << x.s;
	return o;
}

int main()
{
	stable_priority_queue<X> spq;

	spq.push(X("erstes", 1));
	spq.push(X("zweites", 2));
	spq.push(X("drittes", 1));
	spq.push(X("viertes", 2));
	spq.push(X("fuenftes", 1));
	spq.push(X("sechstes", 2));

	while (! spq.empty()) {
		std::cout << spq.top() << std::endl;
		spq.pop();
	}
}
